# MIT License
# Copyright (c) 2026 Franklin Ockerman
# See LICENSE.txt file for full license text

"""Single variant associations.

This module contains functions for calculating the tests statistics used in agricola step 2
"""

from jax import jit, vmap
import jax.numpy as jnp
from jax.numpy.linalg import inv, matrix_rank
from jaxtyping import Array
from jax.scipy.special import expit
from .models import logistic_ridge

### ─────────────────────────────────────────────────────────────
### Helpers
### ─────────────────────────────────────────────────────────────


def ols_block(
    X: Array, L: Array, Y: Array, N_eff: Array, alpha: Array
) -> tuple[Array, Array, Array, Array]:
    XtX = X.T @ X
    XtX_inv = inv(XtX + alpha * jnp.identity(X.shape[1]))
    XtL = X.T @ L
    LtL = L.T @ L
    XtXl_inv22 = inv(LtL - XtL.T @ XtX_inv @ XtL + alpha * jnp.identity(L.shape[1]))
    XtXl_inv12 = -XtX_inv @ XtL @ XtXl_inv22
    XtXl_inv11 = XtX_inv - (XtXl_inv12 @ XtL.T @ XtX_inv)
    XtY = X.T @ Y
    LtY = L.T @ Y
    betax = XtXl_inv11 @ XtY + XtXl_inv12 @ LtY
    betal = XtXl_inv12.T @ XtY + XtXl_inv22 @ LtY
    res = Y - X @ betax - L @ betal
    sse = jnp.sum(res**2, axis=0)
    stat = jnp.einsum("kp,kl,lp", betax, inv(XtXl_inv11), betax) / (
        sse / (N_eff - X.shape[1] - L.shape[1])
    )
    df = matrix_rank(XtX)
    return stat, betax, df, sse


def ols(
    X: Array, Y: Array, N_eff: Array, alpha: Array
) -> tuple[Array, Array, Array, Array]:
    XtX = X.T @ X
    XtY = X.T @ Y
    XtX_inv = inv(XtX + alpha * jnp.identity(X.shape[1]))
    betax = XtX_inv @ XtY
    res = Y - X @ betax
    sse = res.T @ res
    stat = betax.T @ XtX @ betax / (sse / (N_eff - X.shape[1]))
    df = matrix_rank(XtX)
    return stat, betax, df, sse


### ─────────────────────────────────────────────────────────────
### Quantitative Traits
### ─────────────────────────────────────────────────────────────


def _qt_score_lanc(
    G: Array, L: Array, Y: Array, Q: Array, N_eff: Array
) -> tuple[Array, Array, Array, Array, Array]:
    Y = jnp.reshape(Y, Y.shape + (1,) * (2 - Y.ndim))

    K = G.shape[1]
    alpha = N_eff * 1e-6

    ## Genotypes
    H = jnp.sum(G, axis=1)

    ## Residualize by covariates
    G = G - Q @ (Q.T @ G)
    L = L - Q @ (Q.T @ L)
    H = H - Q @ (Q.T @ H)

    LtL_inv = jnp.linalg.inv(L.T @ L + alpha * jnp.identity(K - 1))

    ## Fit null model: Y ~ L
    LtY = L.T @ Y
    r_L = Y - L @ (LtL_inv @ LtY)
    mse_null = jnp.sum(r_L**2, axis=0) / (N_eff - (K - 1))

    ## Score test for anc-deconvoluted genotypes (heterogeneous test)
    LtG = L.T @ G
    Gl = G - L @ (LtL_inv @ LtG)
    U = G.T @ r_L
    GtG = Gl.T @ Gl
    GtG_inv = inv(GtG + alpha * jnp.identity(K))
    chisq_het = jnp.einsum("kp,kl,lp->p", U, GtG_inv, U) / mse_null
    beta_het = U * jnp.diagonal(GtG_inv)[:, None]
    df_het = matrix_rank(GtG)

    ## Score test for genotypes (homogeneous test)
    LtH = L.T @ H
    Hl = H - L @ (LtL_inv @ LtH)
    UH = H.T @ r_L
    HtH = jnp.sum(Hl**2)
    beta_hom = UH / (HtH + alpha)
    chisq_hom = (UH**2) / HtH / mse_null

    return chisq_hom, beta_hom, chisq_het, beta_het, df_het


def _qt_score_nolanc(
    G: Array, Y: Array, Q: Array, N_eff: Array
) -> tuple[Array, Array, Array, Array, Array]:
    Y = jnp.reshape(Y, Y.shape + (1,) * (2 - Y.ndim))

    K = G.shape[1]
    alpha = N_eff * 1e-6

    ## Genotypes
    H = jnp.sum(G, axis=1)

    ## Residualize by covariates
    G = G - Q @ (Q.T @ G)
    H = H - Q @ (Q.T @ H)

    mse_null = jnp.sum(Y**2, axis=0) / N_eff

    ## Score test for anc-deconvoluted genotypes (heterogeneous test)
    U = G.T @ Y
    GtG = G.T @ G
    GtG_inv = inv(GtG + alpha * jnp.identity(K))
    chisq_het = jnp.einsum("kp,kl,lp->p", U, GtG_inv, U) / mse_null
    beta_het = U * jnp.diagonal(GtG_inv)[:, None]
    df_het = matrix_rank(GtG)

    ## Score test for genotypes (homogeneous test)
    UH = H.T @ Y
    HtH = jnp.sum(H**2)
    chisq_hom = (UH**2) / HtH / mse_null
    beta_hom = UH / (HtH + alpha)

    return chisq_hom, beta_hom, chisq_het, beta_het, df_het


def _qt_wald_lanc(
    G: Array, L: Array, Y: Array, Q: Array, N_eff: Array
) -> tuple[Array, Array, Array, Array, Array, Array, Array]:
    Y = jnp.reshape(Y, Y.shape + (1,) * (2 - Y.ndim))

    alpha = N_eff * 1e-6

    ## Genotypes
    H = jnp.sum(G, axis=1, keepdims=True)

    ## Residualize by covariates
    G = G - Q @ (Q.T @ G)
    L = L - Q @ (Q.T @ L)
    H = H - Q @ (Q.T @ H)

    ## Tests
    chisq_het, beta_het, df_het, sse_het = ols_block(G, L, Y, N_eff, alpha)
    chisq_hom, beta_hom, df_hom, sse_hom = ols_block(H, L, Y, N_eff, alpha)
    chisq_lrt = N_eff * jnp.log(sse_hom / sse_het)
    df_lrt = df_het - df_hom

    return chisq_hom, beta_hom[0, :], chisq_het, beta_het, df_het, chisq_lrt, df_lrt


def _qt_wald_nolanc(
    G: Array, Y: Array, Q: Array, N_eff: Array
) -> tuple[Array, Array, Array, Array, Array, Array, Array]:
    Y = jnp.reshape(Y, Y.shape + (1,) * (2 - Y.ndim))

    alpha = N_eff * 1e-6

    ## Genotypes
    H = jnp.sum(G, axis=1, keepdims=True)

    ## Residualize by covariates
    G = G - Q @ (Q.T @ G)
    H = H - Q @ (Q.T @ H)

    ## Tests
    chisq_het, beta_het, df_het, sse_het = ols(G, Y, N_eff, alpha)
    chisq_hom, beta_hom, df_hom, sse_hom = ols(H, Y, N_eff, alpha)
    chisq_lrt = N_eff * jnp.log(sse_hom / sse_het)
    df_lrt = df_het - df_hom

    return chisq_hom, beta_hom[0, :], chisq_het, beta_het, df_het, chisq_lrt, df_lrt


### ─────────────────────────────────────────────────────────────
### Binary Traits
### ─────────────────────────────────────────────────────────────


def _bt_score_lanc(
    G: Array, L: Array, Y: Array, Q: Array, O: Array, M: Array, N_eff: Array
) -> tuple[Array, Array, Array, Array, Array]:
    alpha = N_eff * 1e-6
    K = G.shape[1]

    ## Genotypes
    H = jnp.sum(G, axis=1)

    ## Residualize by covariates
    G = G - Q @ (Q.T @ G)
    L = L - Q @ (Q.T @ L)
    H = H - Q @ (Q.T @ H)

    ## Fit null model L + offset
    beta_L = logistic_ridge(L, Y, O, M, alpha, 10)
    mu = expit(L @ beta_L + O)
    R = (Y - mu) * M
    W_L_sqrt = jnp.sqrt(mu * (1.0 - mu)) * M

    ## Score test for anc-deconvoluted genotypes (heterogeneous test)
    U = G.T @ R
    LtL_inv = jnp.linalg.inv(L.T @ L + alpha * jnp.identity(K - 1))
    LtG = L.T @ G
    Glw = (G - L @ (LtL_inv @ LtG)) * M[:, None] * W_L_sqrt[:, None]
    GtG = Glw.T @ Glw
    GtG_inv = inv(GtG + alpha * jnp.identity(K))
    chisq_het = U.T @ GtG_inv @ U
    beta_het = U * jnp.diagonal(GtG_inv)
    df_het = matrix_rank(GtG)

    ## Score test for genotypes (homogeneous test)
    UH = H.T @ R
    LtH = L.T @ H
    Hlw = (H - L @ (LtL_inv @ LtH)) * M * W_L_sqrt
    HtH = Hlw.T @ Hlw
    beta_hom = UH / (HtH + alpha)
    chisq_hom = (UH**2) / HtH

    return chisq_hom, beta_hom, chisq_het, beta_het, df_het


def _bt_score_nolanc(
    G: Array, Y: Array, Q: Array, O: Array, M: Array, N_eff: Array
) -> tuple[Array, Array, Array, Array, Array]:
    alpha = N_eff * 1e-6
    K = G.shape[1]

    ## Genotypes
    H = jnp.sum(G, axis=1)

    ## Residualize by covariates
    G = G - Q @ (Q.T @ G)
    H = H - Q @ (Q.T @ H)

    mu = expit(O)
    R = (Y - mu) * M
    W_sqrt = jnp.sqrt(mu * (1.0 - mu)) * M

    ## Score test for anc-deconvoluted genotypes (heterogeneous test)
    U = G.T @ R
    Gw = G * M[:, None] * W_sqrt[:, None]
    GtG = Gw.T @ Gw
    GtG_inv = inv(GtG + alpha * jnp.identity(K))
    chisq_het = U.T @ GtG_inv @ U
    beta_het = U * jnp.diagonal(GtG_inv)
    df_het = matrix_rank(GtG)

    ## Score test for genotypes (homogeneous test)
    UH = H.T @ R
    Hw = H * M * W_sqrt
    HtH = Hw.T @ Hw
    beta_hom = UH / (HtH + alpha)
    chisq_hom = (UH**2) / HtH

    return chisq_hom, beta_hom, chisq_het, beta_het, df_het


def _bt_wald_lanc(
    G: Array, L: Array, Y: Array, Q: Array, O: Array, M: Array, N_eff: Array
) -> tuple[Array, Array, Array, Array, Array, Array, Array]:
    alpha = N_eff * 1e-6
    K = G.shape[1]

    ## Genotypes
    H = jnp.sum(G, axis=1)

    ## Residualize by covariates
    G = G - Q @ (Q.T @ G)
    L = L - Q @ (Q.T @ L)
    H = H - Q @ (Q.T @ H)

    ## Wald test for anc-deconvoluted genotypes
    Xg = jnp.concatenate([G, L], axis=1)
    betag = logistic_ridge(Xg, Y, O, M, alpha, 10)
    etag = Xg @ betag + O
    mu = expit(etag)
    W_sqrt = jnp.sqrt(mu * (1 - mu))
    Xw = Xg * W_sqrt[:, None] * M[:, None]
    XtX = Xw.T @ Xw
    XtXw_inv = inv(XtX + alpha * jnp.identity(2 * K - 1))
    chisq_het = betag[:K].T @ inv(XtXw_inv[:K, :K]) @ betag[:K]
    df_het = matrix_rank(XtX[:K, :K])

    ## Wald test for genotypes
    Xh = jnp.concatenate([H[:, None], L], axis=1)
    betah = logistic_ridge(Xh, Y, O, M, alpha, 10)
    etah = Xh @ betah + O
    mu = expit(etah)
    W_sqrt = jnp.sqrt(mu * (1 - mu))
    Xw = Xh * W_sqrt[:, None] * M[:, None]
    XtX = Xw.T @ Xw
    XtXw_inv = inv(XtX + alpha * jnp.identity(K))
    chisq_hom = betah[0] ** 2 / XtXw_inv[0, 0]
    df_hom = matrix_rank(XtX[0, 0])

    ## LRT
    l_het = Y * etag - jnp.log(1 + jnp.exp(etag))
    l_hom = Y * etah - jnp.log(1 + jnp.exp(etah))
    chisq_lrt = 2 * jnp.sum(l_het - l_hom)
    df_lrt = df_het - df_hom

    return chisq_hom, betah[0], chisq_het, betag[:K], df_het, chisq_lrt, df_lrt


def _bt_wald_nolanc(
    G: Array, Y: Array, Q: Array, O: Array, M: Array, N_eff: Array
) -> tuple[Array, Array, Array, Array, Array, Array, Array]:
    alpha = N_eff * 1e-6

    ## Genotypes
    H = jnp.sum(G, axis=1)

    ## Residualize by covariates
    G = G - Q @ (Q.T @ G)
    H = H - Q @ (Q.T @ H)

    ## Wald test for anc-deconvoluted genotypes
    betag = logistic_ridge(G, Y, O, M, alpha, 10)
    etag = G @ betag + O
    mu = expit(etag)
    W_sqrt = jnp.sqrt(mu * (1 - mu))
    Gw = G * W_sqrt[:, None] * M[:, None]
    GtG = Gw.T @ Gw
    chisq_het = betag.T @ GtG @ betag
    df_het = matrix_rank(GtG)

    ## Wald test for genotypes
    betah = logistic_ridge(H[:, None], Y, O, M, alpha, 10)[0]
    etah = H * betah + O
    mu = expit(etah)
    W_sqrt = jnp.sqrt(mu * (1 - mu))
    Hw = H * W_sqrt * M
    HtH = Hw.T @ Hw
    chisq_hom = betah**2 * HtH
    df_hom = matrix_rank(HtH)

    ## LRT
    l_het = Y * etag - jnp.log(1 + jnp.exp(etag))
    l_hom = Y * etah - jnp.log(1 + jnp.exp(etah))
    chisq_lrt = 2 * jnp.sum(l_het - l_hom)
    df_lrt = df_het - df_hom

    return chisq_hom, betah, chisq_het, betag, df_het, chisq_lrt, df_lrt


### ─────────────────────────────────────────────────────────────
### Block-wise functions
### ─────────────────────────────────────────────────────────────

qt_score_lanc = jit(
    vmap(
        vmap(_qt_score_lanc, in_axes=(1, 1, None, None, None)),
        in_axes=(3, 3, 1, 2, 0),
        out_axes=-1,
    )
)

qt_score_lanc_impute = jit(
    vmap(_qt_score_lanc, in_axes=(1, 1, None, None, None)),
)

qt_score_nolanc = jit(
    vmap(
        vmap(_qt_score_nolanc, in_axes=(1, None, None, None)),
        in_axes=(3, 1, 2, 0),
        out_axes=-1,
    )
)

qt_score_nolanc_impute = jit(
    vmap(_qt_score_nolanc, in_axes=(1, None, None, None)),
)

qt_wald_lanc = jit(
    vmap(
        vmap(_qt_wald_lanc, in_axes=(1, 1, None, None, None)),
        in_axes=(3, 3, 1, 2, 0),
        out_axes=-1,
    )
)

qt_wald_lanc_impute = jit(
    vmap(_qt_wald_lanc, in_axes=(1, 1, None, None, None)),
)

qt_wald_nolanc = jit(
    vmap(
        vmap(_qt_wald_nolanc, in_axes=(1, None, None, None)),
        in_axes=(3, 1, 2, 0),
        out_axes=-1,
    )
)

qt_wald_nolanc_impute = jit(
    vmap(_qt_wald_nolanc, in_axes=(1, None, None, None)),
)

bt_score_lanc = jit(
    vmap(
        vmap(_bt_score_lanc, in_axes=(1, 1, None, None, None, None, None)),
        in_axes=(3, 3, 1, 2, 1, 1, 0),
        out_axes=-1,
    )
)

bt_score_nolanc = jit(
    vmap(
        vmap(_bt_score_nolanc, in_axes=(1, None, None, None, None, None)),
        in_axes=(3, 1, 2, 1, 1, 0),
        out_axes=-1,
    )
)


bt_wald_lanc = jit(
    vmap(
        vmap(_bt_wald_lanc, in_axes=(1, 1, None, None, None, None, None)),
        in_axes=(3, 3, 1, 2, 1, 1, 0),
        out_axes=-1,
    )
)

bt_wald_nolanc = jit(
    vmap(
        vmap(_bt_wald_nolanc, in_axes=(1, None, None, None, None, None)),
        in_axes=(3, 1, 2, 1, 1, 0),
        out_axes=-1,
    )
)



# MIT License
# Copyright (c) 2026 Franklin Ockerman
# See LICENSE.txt file for full license text

import os
import pytest
import numpy as np
import jax.numpy as jnp
import jax
from lanctools import LancData
from jax.scipy.special import expit
from agricola._internal.level0 import level0
from agricola._internal.level1 import level1


@pytest.fixture
def toy_data():
    data = [
        LancData(
            plink_prefix="tests/data/chr" + str(chr),
            lanc_file="tests/data/chr" + str(chr) + ".lanc",
        )
        for chr in range(20, 23)
    ]
    return data


def valid_inputs_0(P=3, K=5, C=1, H=4):
    """Utility for constructing minimal valid inputs for step 0."""
    N = 20
    key0 = jax.random.key(30091324)
    key = jax.random.key(8899134)
    Y = jax.random.normal(key0, shape=(N, P))
    X = jax.random.normal(key, shape=(N, C))
    train_mask = jnp.ones((N, K))
    test_mask = jnp.ones((N, K))
    h2_prior = jnp.linspace(0.1, 0.9, H)
    phenotypes = [str(i) for i in range(Y.shape[1])]
    return Y, X, phenotypes, train_mask, test_mask, h2_prior


def valid_inputs_1(P=3, C=1, H=4):
    """Utility for constructing minimal valid inputs."""
    pref = "tests/data/level0/"
    level0_dicts = [
        {
            "20": f"{pref}{i}_20.npy",
            "21": f"{pref}{i}_21.npy",
            "22": f"{pref}{i}_22.npy",
        }
        for i in range(3)
    ]
    level0_files = dict(zip([str(i) for i in range(3)], level0_dicts))
    Y, X, phenotypes, train_mask, test_mask, h2_prior = valid_inputs_0(P, C, H)
    h2_prior = jnp.linspace(0.1, 0.9, H)
    return level0_files, Y, X, phenotypes, train_mask, test_mask, h2_prior


### ─────────────────────────────────────────────────────────────
### Level 0
### ─────────────────────────────────────────────────────────────


def test_step0_dataset_elements_type_error():
    """Check that bad datasets throws error"""
    Y, X, phenotypes, train_mask, test_mask, h2 = valid_inputs_0()
    with pytest.raises(TypeError, match="must be LancData"):
        level0([object()], Y, X, phenotypes, train_mask, test_mask, h2)  # pyright: ignore


def test_level0_Y_dim_error(
    toy_data,
):
    """Check that 1D Y throws error"""
    _, X, phenotypes, train_mask, test_mask, h2 = valid_inputs_0()
    with pytest.raises(ValueError, match="Y must be 2D"):
        level0(toy_data, jnp.zeros(5), X, phenotypes, train_mask, test_mask, h2)


def test_level0_X_dim_error(
    toy_data,
):
    """Check that 1D X throws error"""
    Y, _, phenotypes, train_mask, test_mask, h2 = valid_inputs_0()
    with pytest.raises(ValueError, match="X must be 2D"):
        level0(toy_data, Y, jnp.zeros((10,)), phenotypes, train_mask, test_mask, h2)


def test_level0_X_matches_N_error(
    toy_data,
):
    """Check that N mis-match with X throws error"""
    Y, _, phenotypes, train_mask, test_mask, h2 = valid_inputs_0()
    wrong_X = jnp.zeros((5, 2))
    with pytest.raises(ValueError, match="must match Y.shape"):
        level0(toy_data, Y, wrong_X, phenotypes, train_mask, test_mask, h2)


def test_level0_h2_prior_dim_error(
    toy_data,
):
    """Check that h2_prior with wrong dimension throws error"""
    Y, X, phenotypes, train_mask, test_mask, _ = valid_inputs_0()
    with pytest.raises(ValueError, match="h2_prior must be 1D"):
        level0(toy_data, Y, X, phenotypes, train_mask, test_mask, jnp.ones((3, 2)))


def test_level0_h2_prior_domain_error(
    toy_data,
):
    """Check that h2_prior outside (0,1) throws error"""
    Y, X, phenotypes, train_mask, test_mask, _ = valid_inputs_0()
    with pytest.raises(ValueError, match="in the open interval"):
        level0(
            toy_data,
            Y,
            X,
            phenotypes,
            train_mask,
            test_mask,
            jnp.array([0.5, 0.0, 0.7]),
        )


@pytest.mark.parametrize("B", [0, -5, 3.14, "foo"])
def test_level0_B_error(
    B,
    toy_data,
):
    """Check that bad B throws error"""
    Y, X, phenotypes, train_mask, test_mask, h2 = valid_inputs_0()
    with pytest.raises(ValueError, match="B must be a positive integer"):
        level0(toy_data, Y, X, phenotypes, train_mask, test_mask, h2, B=B)


def test_level0_variants_type_error(
    toy_data,
):
    """Check that bad variants type throws error"""
    Y, X, phenotypes, train_mask, test_mask, h2 = valid_inputs_0()
    with pytest.raises(TypeError, match="variants must be a list of strings"):
        level0(
            toy_data,
            Y,
            X,
            phenotypes,
            train_mask,
            test_mask,
            h2,
            variants=[1, 2, 3],  # pyright: ignore
        )


def test_level0_validation_happy_path(
    toy_data,
):
    """Check that valid data throws no type errors"""
    Y, X, phenotypes, train_mask, test_mask, h2 = valid_inputs_0()
    level0(toy_data, Y, X, phenotypes, train_mask, test_mask, h2, B=100)


### ─────────────────────────────────────────────────────────────
### Level 1
### ─────────────────────────────────────────────────────────────


def test_level1_Y_dim_error():
    """Check that 1D Y throws error"""
    level0_files, _, X, phenos, train, test, h2 = valid_inputs_1()
    with pytest.raises(ValueError, match="Y must be 2D"):
        level1(level0_files, jnp.zeros(5), X, phenos, train, test, h2, "qt")

    with pytest.raises(ValueError, match="Y must be 2D"):
        level1(level0_files, jnp.zeros(5), X, phenos, train, test, h2, "bt")


def test_level1_X_dim_error():
    """Check that 1D X throws error"""
    level0_files, Y, _, phenos, train, test, h2 = valid_inputs_1()
    with pytest.raises(ValueError, match="X must be 2D"):
        level1(level0_files, Y, jnp.zeros((10,)), phenos, train, test, h2, "qt")

    Y = jnp.round(expit(Y))
    with pytest.raises(ValueError, match="X must be 2D"):
        level1(level0_files, Y, jnp.zeros((10,)), phenos, train, test, h2, "bt")


def test_level1_X_matches_N_error():
    """Check that N mis-match with X throws error"""
    level0_files, Y, _, phenos, train, test, h2 = valid_inputs_1()
    with pytest.raises(ValueError, match="must match Y.shape"):
        level1(level0_files, Y, jnp.zeros((5, 2)), phenos, train, test, h2, "qt")

    Y = jnp.round(expit(Y))
    with pytest.raises(ValueError, match="must match Y.shape"):
        level1(level0_files, Y, jnp.zeros((5, 2)), phenos, train, test, h2, "bt")


def test_level1_mask_dim_error():
    """Check that 1D train/test mask throws error"""
    level0_files, Y, X, phenos, _, _, h2 = valid_inputs_1()
    with pytest.raises(ValueError, match="train_mask and test_mask must be 2D"):
        level1(level0_files, Y, X, phenos, jnp.ones((10,)), jnp.ones((10, 2)), h2, "qt")

    Y = jnp.round(expit(Y))
    with pytest.raises(ValueError, match="train_mask and test_mask must be 2D"):
        level1(level0_files, Y, X, phenos, jnp.ones((10,)), jnp.ones((10, 2)), h2, "bt")


def test_level1_mask_shape_mismatch_error():
    """Check that train/test mask shape mis-match throws error"""
    level0_files, Y, X, phenos, _, _, h2 = valid_inputs_1()
    with pytest.raises(ValueError, match="same shape"):
        level1(
            level0_files, Y, X, phenos, jnp.ones((10, 2)), jnp.ones((10, 3)), h2, "qt"
        )

    Y = jnp.round(expit(Y))
    with pytest.raises(ValueError, match="same shape"):
        level1(
            level0_files, Y, X, phenos, jnp.ones((10, 2)), jnp.ones((10, 3)), h2, "bt"
        )


def test_level1_h2_prior_dim_error():
    """Check that h2_prior with wrong dimension throws error"""
    level0_files, Y, X, phenos, train, test, _ = valid_inputs_1()
    with pytest.raises(ValueError, match="h2_prior must be 1D"):
        level1(level0_files, Y, X, phenos, train, test, jnp.ones((3, 2)), "qt")

    Y = jnp.round(expit(Y))
    with pytest.raises(ValueError, match="h2_prior must be 1D"):
        level1(level0_files, Y, X, phenos, train, test, jnp.ones((3, 2)), "bt")


def test_level1_h2_prior_domain_error():
    """Check that h2_prior outside (0,1) throws error"""
    level0_files, Y, X, phenos, train, test, _ = valid_inputs_1()
    with pytest.raises(ValueError, match="in the open interval"):
        level1(
            level0_files, Y, X, phenos, train, test, jnp.array([0.5, 0.0, 0.7]), "qt"
        )

    Y = jnp.round(expit(Y))
    with pytest.raises(ValueError, match="in the open interval"):
        level1(
            level0_files, Y, X, phenos, train, test, jnp.array([0.5, 0.0, 0.7]), "bt"
        )


def test_level1_qt_valid_input():
    """Check that valid data throws no type errors"""
    level0_files, Y, X, phenos, train, test, h2 = valid_inputs_1()
    level1(level0_files, Y, X, phenos, train, test, h2, "qt")


def test_level1_bt_valid_input():
    """Check that valid data throws no type errors"""
    ## cross-validation
    level0_files, Y, X, phenos, train, test, h2 = valid_inputs_1()
    Y = jnp.round(expit(Y))
    level1(level0_files, Y, X, phenos, train, test, h2, "bt", False)

    ## loco
    level0_files, Y, X, phenos, train, test, h2 = valid_inputs_1()
    level1(level0_files, Y, X, phenos, train, test, h2, "bt", True)



# MIT License
# Copyright (c) 2026 Franklin Ockerman
# See LICENSE.txt file for full license text

import pytest
import numpy as np
import jax.numpy as jnp
import jax
from jax.scipy.special import expit
import pandas as pd
from agricola import step2
from lanctools import LancData


@pytest.fixture
def toy_data():
    data = [
        LancData(
            plink_prefix="tests/data/chr" + str(chr),
            lanc_file="tests/data/chr" + str(chr) + ".lanc",
        )
        for chr in range(20, 23)
    ]
    return data


def valid_inputs(P=3, C=1):
    """Utility for constructing minimal valid inputs."""
    key0 = jax.random.key(8899134)
    key1 = jax.random.key(471464)
    key2 = jax.random.key(7314120)
    N = 20
    step1_predictions = {}
    for i in range(20, 23):
        arr = jax.random.normal(jax.random.split(key0, 3)[i - 21], shape=(N, P))
        df = pd.DataFrame(arr, columns=pd.Index([str(i) for i in range(3)]))
        step1_predictions[str(i)] = df

    Y = jax.random.normal(key1, shape=(N, P))
    X = jax.random.normal(key2, shape=(N, C))
    return Y, X, step1_predictions


def test_step2_dataset_elements_type_error(tmp_path):
    """Check that bad datasets throws error"""
    Y, X, step1_predictions = valid_inputs()
    phenotypes = [str(i) for i in range(3)]
    out_prefixes = [tmp_path / f"_{i}" for i in range(20, 23)]
    with pytest.raises(TypeError, match="must be LancData"):
        step2([object()], Y, X, step1_predictions, out_prefixes, phenotypes, "qt")  # pyright: ignore

    Y = jnp.round(expit(Y))
    with pytest.raises(TypeError, match="must be LancData"):
        step2([object()], Y, X, step1_predictions, out_prefixes, phenotypes, "bt")  # pyright: ignore


def test_step2_pred_dim_error(tmp_path):
    """Check that bad Z dimensions throws error"""
    Y, X, step1_predictions = valid_inputs()
    step1_predictions["21"] = np.ones(shape=(1, 6, 4))
    phenotypes = [i for i in range(3)]
    out_prefixes = [tmp_path / f"_{i}" for i in range(20, 23)]
    with pytest.raises(TypeError, match="must be LancData"):
        step2([object()], Y, X, step1_predictions, out_prefixes, phenotypes, "qt")  # pyright: ignore

    Y = jnp.round(expit(Y))
    with pytest.raises(TypeError, match="must be LancData"):
        step2([object()], Y, X, step1_predictions, out_prefixes, phenotypes, "bt")  # pyright: ignore


def test_step2_Y_dim_error(tmp_path, toy_data):
    """Check that 1D Y throws error"""
    _, X, step1_predictions = valid_inputs()
    phenotypes = [str(i) for i in range(3)]
    out_prefixes = [tmp_path / f"_{i}" for i in range(20, 23)]
    with pytest.raises(ValueError, match="Y must be 2D"):
        step2(
            toy_data, jnp.zeros(5), X, step1_predictions, out_prefixes, phenotypes, "qt"
        )  # pyright: ignore

    with pytest.raises(ValueError, match="Y must be 2D"):
        step2(
            toy_data, jnp.zeros(5), X, step1_predictions, out_prefixes, phenotypes, "bt"
        )  # pyright: ignore


def test_step2_X_dim_error(tmp_path, toy_data):
    """Check that 1D X throws error"""
    Y, _, step1_predictions = valid_inputs()
    phenotypes = [str(i) for i in range(3)]
    out_prefixes = [tmp_path / f"_{i}" for i in range(20, 23)]
    with pytest.raises(ValueError, match="X must be 2D"):
        step2(
            toy_data,
            Y,
            jnp.zeros(10),
            step1_predictions,
            out_prefixes,
            phenotypes,
            "qt",
        )  # pyright: ignore

    Y = jnp.round(expit(Y))
    with pytest.raises(ValueError, match="X must be 2D"):
        step2(
            toy_data,
            Y,
            jnp.zeros(10),
            step1_predictions,
            out_prefixes,
            phenotypes,
            "bt",
        )  # pyright: ignore


def test_step2_X_matches_N_error(tmp_path, toy_data):
    """Check that N mis-match with X throws error"""
    Y, X, step1_predictions = valid_inputs()
    phenotypes = [str(i) for i in range(3)]
    out_prefixes = [tmp_path / f"_{i}" for i in range(20, 23)]
    with pytest.raises(ValueError, match="must match Y.shape"):
        step2(
            toy_data,
            Y,
            jnp.zeros((5, 2)),
            step1_predictions,
            out_prefixes,
            phenotypes,
            "qt",
        )  # pyright: ignore

    Y = jnp.round(expit(Y))
    with pytest.raises(ValueError, match="must match Y.shape"):
        step2(
            toy_data,
            Y,
            jnp.zeros((5, 2)),
            step1_predictions,
            out_prefixes,
            phenotypes,
            "bt",
        )  # pyright: ignore


### ─────────────────────────────────────────────────────────────
### Other
### ─────────────────────────────────────────────────────────────


def test_step2_valid_input(tmp_path, toy_data):
    """Check that valid data throws no type errors"""
    Y, X, step1_predictions = valid_inputs()
    phenotypes = [str(i) for i in range(3)]
    out_prefixes = [tmp_path / f"_{i}" for i in range(20, 23)]
    step2(toy_data, Y, X, step1_predictions, out_prefixes, phenotypes, "qt")
    step2(
        toy_data,
        Y,
        X,
        step1_predictions,
        out_prefixes,
        phenotypes,
        "qt",
        adjust_lanc=True,
    )

    Y = jnp.round(expit(Y))
    step2(toy_data, Y, X, step1_predictions, out_prefixes, phenotypes, "bt")
    step2(
        toy_data,
        Y,
        X,
        step1_predictions,
        out_prefixes,
        phenotypes,
        "bt",
        adjust_lanc=True,
    )



# MIT License
# Copyright (c) 2026 Franklin Ockerman
# See LICENSE.txt file for full license text

"""Functions and classes used by multiple modules"""

import jax
from jaxtyping import Array
import jax.numpy as jnp
from lanctools import LancData
import numpy as np
from numpy.typing import NDArray
from enum import Enum

### ─────────────────────────────────────────────────────────────
### Classes
### ─────────────────────────────────────────────────────────────


class TestType(Enum):
    SCORE = "score"
    WALD = "wald"


class TraitType(Enum):
    QT = "qt"
    BT = "bt"


### ─────────────────────────────────────────────────────────────
### Functions
### ─────────────────────────────────────────────────────────────


def stdize(X: Array) -> Array:
    """Safely standardizes an ndarray along the first dimension.

    For columns with standard deviation of 0, this function does NOT subtract
    the mean. This behavior is intended to preserve an intercept, for example.
    """
    mean = X.mean(axis=0, keepdims=True)
    std = X.std(axis=0, keepdims=True)
    mean = jnp.where(std == 0, 0, mean)
    std = jnp.where(std == 0, 1, std)
    result = (X - mean) / std
    return result


def get_cv_mask(n: int, k: int, key: Array) -> tuple[Array, Array]:
    """Generate boolean train/test masks for k-fold cross-validation."""
    idx = jax.random.permutation(key, n)
    fold_sizes = jnp.full(k, n // k)
    fold_sizes = fold_sizes.at[: n % k].add(1)
    folds = []
    start = 0
    for size in fold_sizes:
        folds.append(idx[start : start + size])
        start += size

    train_mask = jnp.zeros((n, k), dtype=bool)
    test_mask = jnp.zeros((n, k), dtype=bool)

    for fold in range(k):
        test_idx = folds[fold]
        train_idx = jnp.concatenate([folds[i] for i in range(k) if i != fold])
        test_mask = test_mask.at[test_idx, fold].set(True)
        train_mask = train_mask.at[train_idx, fold].set(True)

    return train_mask, test_mask


def assert_covar_full_rank(X: Array, rtol: float = 1e-8) -> None:
    """Raises ValueError if X does not have full column rank.

    Assumes X is standardized (columns ~ unit variance).
    """
    S = jnp.linalg.svd(X, compute_uv=False)
    rank = jnp.sum(S > rtol * S[0])

    if rank < X.shape[1]:
        raise ValueError(
            f"Collinearity detected in : rank={rank}, n_features={X.shape[1]}"
        )


def get_geno_lanc_deconv(
    dataset: LancData, indices: NDArray[np.uint32]
) -> tuple[Array, Array]:
    """Queries ancestry-masked genotypes and local ancestries.

    Args:
        dataset: A LancData object
        indices: A (V,) jax array of (ascending ordered) variant indices

    Returns:
        A tuple consisting of:
            1) geno_masked: an (N, V, K) jax array of genotypes masked by K
            local ancestries. E.g. geno_masked[:,0,1] = sum of haplotypes with
            ancestry 1 and an alternative allele at variant index 0.
            2) lanc_masked: an (N, V, K) jax array of local ancestries.
            E.g. lanc_masked[:,0,1] = sum of haplotypes with ancestry 1
            at variant index 0.
    """
    geno = jnp.asarray(dataset.get_geno(indices))
    lanc = jnp.asarray(dataset.get_lanc(indices))
    ancestries = jnp.arange(len(dataset.ancestries))
    left_haps_mask = (lanc[:, :, 0:1] == ancestries[None, None, :]).astype(np.int32)
    right_haps_mask = (lanc[:, :, 1:2] == ancestries[None, None, :]).astype(np.int32)
    geno_masked = left_haps_mask * geno[:, :, 0:1] + right_haps_mask * geno[:, :, 1:2]
    lanc_masked = left_haps_mask + right_haps_mask
    return geno_masked, lanc_masked


def get_geno_deconv(dataset: LancData, indices: NDArray[np.uint32]) -> Array:
    """Queries ancestry-masked genotypes.

    Args:
        dataset: A LancData object
        indices: A (V,) jax array of (ascending ordered) variant indices

    Returns:
        A (N, V, K) jax array of genotypes masked by K local ancestries. E.g.
            geno_masked[:,0,1] = sum of haplotypes with ancestry 1 and an
            alternative allele at variant index 0.
    """
    geno = jnp.asarray(dataset.get_geno(indices))
    lanc = jnp.asarray(dataset.get_lanc(indices))
    ancestries = jnp.arange(len(dataset.ancestries))
    left_haps_mask = (lanc[:, :, 0:1] == ancestries[None, None, :]).astype(np.int32)
    right_haps_mask = (lanc[:, :, 1:2] == ancestries[None, None, :]).astype(np.int32)
    geno_masked = left_haps_mask * geno[:, :, 0:1] + right_haps_mask * geno[:, :, 1:2]
    return geno_masked
